@@ -22,13 +22,13 @@
 #include <config.h>
 #include "workspace.h"
 #include "errors.h"
+#include "prefs.h"
 #include <X11/Xatom.h>
 #include <string.h>
 
 void meta_workspace_queue_calc_showing  (MetaWorkspace *workspace);
 
-static int set_number_of_spaces_hint  (MetaScreen *screen);
-static int set_active_space_hint      (MetaScreen *screen);
+static void set_active_space_hint      (MetaScreen *screen);
 
 MetaWorkspace*
 meta_workspace_new (MetaScreen *screen)
@@ -38,8 +38,8 @@ meta_workspace_new (MetaScreen *screen)
   workspace = g_new (MetaWorkspace, 1);
 
   workspace->screen = screen;
-  workspace->screen->display->workspaces =
-    g_list_append (workspace->screen->display->workspaces, workspace);
+  workspace->screen->workspaces =
+    g_list_append (workspace->screen->workspaces, workspace);
   workspace->windows = NULL;
 
   workspace->work_area.x = 0;
@@ -47,12 +47,6 @@ meta_workspace_new (MetaScreen *screen)
   workspace->work_area.width = screen->width;
   workspace->work_area.height = screen->height;
   workspace->work_area_invalid = TRUE;
-
-  workspace->name = g_strdup_printf (_("Workspace %d"),
-                                     meta_workspace_index (workspace) + 1);
-  
-  /* Update hint for current number of workspaces */
-  set_number_of_spaces_hint (screen);
   
   return workspace;
 }
@@ -87,15 +81,11 @@ meta_workspace_free (MetaWorkspace *workspace)
 
   screen = workspace->screen;
   
-  workspace->screen->display->workspaces =
-    g_list_remove (workspace->screen->display->workspaces, workspace);
-
-  g_free (workspace->name);
+  workspace->screen->workspaces =
+    g_list_remove (workspace->screen->workspaces, workspace);
   
   g_free (workspace);
 
-  /* Update hint for current number of workspaces */
-  set_number_of_spaces_hint (screen);
   /* don't bother to reset names, pagers can just ignore
    * extra ones
    */
@@ -240,7 +230,7 @@ meta_workspace_index (MetaWorkspace *workspace)
   int i;
 
   i = 0;
-  tmp = workspace->screen->display->workspaces;
+  tmp = workspace->screen->workspaces;
   while (tmp != NULL)
     {
       if (tmp->data == workspace)
@@ -252,30 +242,7 @@ meta_workspace_index (MetaWorkspace *workspace)
     }
 
   meta_bug ("Workspace does not exist to index!\n");
-}
-
-int
-meta_workspace_screen_index  (MetaWorkspace *workspace)
-{
-  GList *tmp;
-  int i;
-
-  i = 0;
-  tmp = workspace->screen->display->workspaces;
-  while (tmp != NULL)
-    {
-      MetaWorkspace *w = tmp->data;
-
-      if (tmp->data == workspace)
-        return i;
-
-      if (w->screen == workspace->screen)
-        ++i;
-                    
-      tmp = tmp->next;
-    }
-
-  meta_bug ("Workspace does not exist to index!\n");
+  return -1; /* compiler warnings */
 }
 
 /* get windows contained on workspace, including workspace->windows
@@ -308,32 +275,21 @@ meta_workspace_list_windows (MetaWorkspace *workspace)
   return workspace_windows;
 }
 
-static int
-set_number_of_spaces_hint (MetaScreen *screen)
-{
-  unsigned long data[1];
-
-  if (screen->closing > 0)
-    return 0;
-  
-  data[0] = meta_screen_get_n_workspaces (screen);
-
-  meta_verbose ("Setting _NET_NUMBER_OF_DESKTOPS to %ld\n", data[0]);
-  
-  meta_error_trap_push (screen->display);
-  XChangeProperty (screen->display->xdisplay, screen->xroot,
-                   screen->display->atom_net_number_of_desktops,
-                   XA_CARDINAL,
-                   32, PropModeReplace, (guchar*) data, 1);
-  return meta_error_trap_pop (screen->display);
-}
-
-static int
+static void
 set_active_space_hint (MetaScreen *screen)
 {
   unsigned long data[1];
+
+  /* this is because we destroy the spaces in order,
+   * so we always end up setting a current desktop of
+   * 0 when closing a screen, so lose the current desktop
+   * on restart. By doing this we keep the current
+   * desktop on restart.
+   */
+  if (screen->closing > 0)
+    return;
   
-  data[0] = meta_workspace_screen_index (screen->active_workspace);
+  data[0] = meta_workspace_index (screen->active_workspace);
 
   meta_verbose ("Setting _NET_CURRENT_DESKTOP to %ld\n", data[0]);
   
@@ -342,7 +298,7 @@ set_active_space_hint (MetaScreen *screen)
                    screen->display->atom_net_current_desktop,
                    XA_CARDINAL,
                    32, PropModeReplace, (guchar*) data, 1);
-  return meta_error_trap_pop (screen->display);
+  meta_error_trap_pop (screen->display, FALSE);
 }
 
 void
@@ -459,124 +415,190 @@ meta_workspace_get_work_area (MetaWorkspace *workspace,
   *area = workspace->work_area;
 }
 
+static char *
+meta_motion_direction_to_string (MetaMotionDirection direction)
+{
+  switch (direction)
+    {
+    case META_MOTION_UP:
+      return "Up";
+    case META_MOTION_DOWN:
+      return "Down";
+    case META_MOTION_LEFT:
+      return "Left";
+    case META_MOTION_RIGHT:
+      return "Right";
+    }
+
+  return "Unknown";
+}
+
+static char *
+meta_screen_corner_to_string (MetaScreenCorner corner)
+{
+  switch (corner)
+    {
+    case META_SCREEN_TOPLEFT:
+      return "TopLeft";
+    case META_SCREEN_TOPRIGHT:
+      return "TopRight";
+    case META_SCREEN_BOTTOMLEFT:
+      return "BottomLeft";
+    case META_SCREEN_BOTTOMRIGHT:
+      return "BottomRight";
+    }
+
+  return "Unknown";
+}
+
 MetaWorkspace*
 meta_workspace_get_neighbor (MetaWorkspace      *workspace,
                              MetaMotionDirection direction)
 {
   int i, num_workspaces, grid_area;
   int rows, cols;
+  int new_workspace_idx;
+  int up_diff, down_diff, left_diff, right_diff; 
+  int current_row, current_col;
   
   i = meta_workspace_index (workspace);
   num_workspaces = meta_screen_get_n_workspaces (workspace->screen);
   
- meta_screen_calc_workspace_layout (workspace->screen, num_workspaces,
-                                    &rows, &cols);
+  meta_screen_calc_workspace_layout (workspace->screen, num_workspaces,
+                                     &rows, &cols);
+
+  g_assert (rows != 0 && cols != 0);
 
   grid_area = rows * cols;
-  
-  meta_verbose ("Getting neighbor rows = %d cols = %d vert = %d "
-                "current = %d num_spaces = %d neighbor = %d\n",
-                rows, cols, workspace->screen->vertical_workspaces,
-                i, num_workspaces, direction);
 
+  meta_verbose ("Getting neighbor rows = %d cols = %d current = %d "
+                "num_spaces = %d vertical = %s direction = %s corner = %s\n",
+                rows, cols, i, num_workspaces,
+                workspace->screen->vertical_workspaces ? "(true)" : "(false)",
+	        meta_motion_direction_to_string (direction),
+                meta_screen_corner_to_string (workspace->screen->starting_corner));
   
-  /* FIXME someone who's better or more patient at math puzzles
-   * should clean this up. ;-)
+  /* ok, we want to setup the distances in the workspace array to go     
+   * in each direction. Remember, there are many ways that a workspace   
+   * array can be setup.                                                 
+   * see http://www.freedesktop.org/standards/wm-spec/1.2/html/x109.html 
+   * and look at the _NET_DESKTOP_LAYOUT section for details.            
+   * For instance:
    */
-  if (workspace->screen->vertical_workspaces)
+  /* starting_corner = META_SCREEN_TOPLEFT                         
+   *  vertical_workspaces = 0                 vertical_workspaces=1
+   *       1234                                    1357            
+   *       5678                                    2468            
+   *                                                               
+   * starting_corner = META_SCREEN_TOPRIGHT                        
+   *  vertical_workspaces = 0                 vertical_workspaces=1
+   *       4321                                    7531            
+   *       8765                                    8642            
+   *                                                               
+   * starting_corner = META_SCREEN_BOTTOMLEFT                      
+   *  vertical_workspaces = 0                 vertical_workspaces=1
+   *       5678                                    2468            
+   *       1234                                    1357            
+   *                                                               
+   * starting_corner = META_SCREEN_BOTTOMRIGHT                     
+   *  vertical_workspaces = 0                 vertical_workspaces=1
+   *       8765                                    8642            
+   *       4321                                    7531            
+   *
+   */
+
+  if (workspace->screen->vertical_workspaces) 
     {
-      switch (direction)
-        {
-        case META_MOTION_LEFT:
-          if (i < rows)
-            {
-              i = grid_area - (i % rows) - 1;
-              while (i >= num_workspaces)
-                i -= rows;
-            }
-          else
-            i -= rows;
-          break;
-        case META_MOTION_RIGHT:
-          if ((i + rows) >= num_workspaces)
-            i = (i % rows + 1) % rows;
-          else
-            i += rows;
-          break;
-        case META_MOTION_UP:
-          --i;
-          if (i == -1)
-            i = num_workspaces - 1;
-          break;
-        case META_MOTION_DOWN:
-          ++i;
-          if (i == num_workspaces)
-            i = 0;
-          break;
-        }
+      up_diff    = -1;
+      down_diff  = 1;
+      left_diff  = -1 * rows;
+      right_diff = rows;
+      current_col = ((i - 1) / rows) + 1;
+      current_row = ((i - 1) % rows) + 1;
     }
-  else
+  else 
     {
-      switch (direction)
-        {
-        case META_MOTION_LEFT:
-          --i;
-          if (i == -1)
-            i = num_workspaces - 1;
-          break;
-        case META_MOTION_RIGHT:
-          ++i;
-          if (i == num_workspaces)
-            i = 0;
-          break;
-        case META_MOTION_UP:
-          if (i < cols)
-            {
-	      if (i == 0)
-		{
-		  /* special case: go to bottom right corner == grid_area - 1 */
-		  i = grid_area - 1;
-		  while (i >= num_workspaces)
-		    /* if we are in a non-existant workspace look up in the column for next workspace */
-		    i -= cols;
-		}
-	      else
-		{
-		  /* calculate new workspace based on the bottom left corner == ((rows - 1) * cols) */
-		  i = ((rows - 1) * cols) + (i - 1);
-		  while (i >= num_workspaces)
-		    /* if we are in a non-existant workspace look up in the column for next workspace */
-		    i -= cols;
-		}
-            }
-          else
-            i -= cols;
-          break;
-        case META_MOTION_DOWN:
-          if ((i + cols) >= num_workspaces)
-            i = (i % cols + 1) % cols;
-          else
-            i += cols;
-          break;
-        }
+      up_diff    = -1 * cols;
+      down_diff  = cols;
+      left_diff  = -1;
+      right_diff = 1;
+      current_col = (i % cols) + 1;
+      current_row = ((i - 1) / cols) + 1;
     }
 
-  meta_verbose ("Neighbor space is %d\n", i);
+  switch (workspace->screen->starting_corner) 
+    {
+    default:
+    case META_SCREEN_TOPLEFT:
+      /* this was the default case setup in the if() above */
+      break;
+    case META_SCREEN_TOPRIGHT:
+      /* ok, we need to inverse the left/right values      */
+      left_diff  = -1 * left_diff;
+      right_diff = -1 * right_diff;
+      /* also, current column needs to be mirrored         */
+      current_col = rows - ((current_col-1)%rows) ;
+      break;
+    case META_SCREEN_BOTTOMLEFT:
+      /* ok, we need to inverse the up/down values         */
+      up_diff    = -1 * up_diff;
+      down_diff  = -1 * up_diff;
+      /* also, current row needs to be mirrored            */
+      current_row = cols - ((current_row-1)%cols);
+      break;
+    case META_SCREEN_BOTTOMRIGHT:
+      /* in this case, we need to inverse everything       */
+      up_diff    = -1 * up_diff;
+      down_diff  = -1 * up_diff;
+      left_diff  = -1 * left_diff;
+      right_diff = -1 * right_diff;
+      /* also, current column and row need to be reversed  */
+      current_col = rows - ((current_col-1)%rows);
+      current_row = cols - ((current_row-1)%cols);
+      break;
+    }
+
+  meta_verbose ("Workspace deltas: up = %d down = %d left = %d right = %d. "
+                "Current col = %d row = %d\n", up_diff, down_diff, left_diff,
+                right_diff, current_col, current_row);
+
+  /* calculate what we think the next spot should be */
+  new_workspace_idx = i;
+
+  switch (direction) 
+    {
+    case META_MOTION_LEFT:
+      if (current_col >= 1)
+        new_workspace_idx = i + left_diff;
+      break;
+    case META_MOTION_RIGHT:
+      if (current_col <= cols)
+        new_workspace_idx = i + right_diff;
+      break;
+    case META_MOTION_UP:
+      if (current_row >= 1)
+        new_workspace_idx = i + up_diff;
+      break;
+    case META_MOTION_DOWN:
+      if (current_row <= rows)
+        new_workspace_idx = i + down_diff;
+      break;
+    default:
+      new_workspace_idx = 0;
+      break;
+    }
+
+  /* and now make sure we don't over/under flow */
+  if ((new_workspace_idx >= 0) && (new_workspace_idx < num_workspaces)) 
+    i = new_workspace_idx;
   
-  return meta_display_get_workspace_by_index (workspace->screen->display,
-                                              i);
+  meta_verbose ("Neighbor workspace is %d\n", i);
+  
+  return meta_screen_get_workspace_by_index (workspace->screen, i);
 }
 
-void
-meta_workspace_set_name (MetaWorkspace *workspace,
-                         const char    *name)
+const char*
+meta_workspace_get_name (MetaWorkspace *workspace)
 {
-  if (strcmp (name, workspace->name) == 0)
-    return;
-  
-  g_free (workspace->name);
-  workspace->name = g_strdup (name);
-
-  meta_verbose ("Workspace has new name \"%s\"\n",
-                workspace->name);
+  return meta_prefs_get_workspace_name (meta_workspace_index (workspace));
 }
